#include <CheckUp.h>


CheckUp::CheckUp()
{}

CheckUp::~CheckUp()
{}




/*!
     \brief Check each item of the robot
     \brief if an error is detected, wait for the push button
     \return 1 : check up completed
     |return 0 : fatal error
  */
void CheckUp::Robot_CheckUp(MiniRex* Robot)
{

    this->Check_Batteries(Robot);
    this->Check_PushButton(Robot);
    this->Check_Inclinometer(Robot);
    this->Check_Kinect(Robot);
    this->Check_Servomotor(Robot);
    this->Check_Tracks(Robot);
    //this->Check_UltraSonic(Robot);        // A decommenter des que les capteurs sont montes
    this->Check_DiskFreeSpace(Robot);
    this->Check_DataDirectory(Robot);
    fflush(stdout);
    Robot->Kinect.SetLedOption(LED_BLINK_GREEN);
}




/*!
     \brief Check batteries voltage
     \return None
  */
void CheckUp::Check_Batteries(MiniRex* Robot)
{
    unsigned short BatElec,BatPower;
    char Buffer[128];
    char Ret=Robot->Firmware.GetBatteryVoltage_mV(&BatElec,&BatPower);
    if (Ret!=1) {
        this->DisplayErrorMessageAndValid(Robot,"Batteries voltage, error while requesting data");
        return;
    }
    if (BatElec<20000 || BatElec>25000)
    {
        sprintf (Buffer,"Batteries [electronics] %5.2fV",BatElec/1000.0);
        this->DisplayErrorMessageAndValid(Robot, Buffer);
    }
    else {
        sprintf (Buffer,"Batteries [electronics] %5.2fV",BatElec/1000.0);
        this->DisplaySuccessMessage(Buffer);
    }

    if (BatPower<20000 || BatPower>25000)
    {
        sprintf (Buffer,"Batteries [Power] %5.2fV",BatPower/1000.0);
        this->DisplayErrorMessageAndValid(Robot, Buffer);
    }
    else {
        sprintf (Buffer,"Batteries [Power] %5.2fV",BatPower/1000.0);
        this->DisplaySuccessMessage(Buffer);
    }
    fflush(stdout);
}




/*!
\brief Check push-button
\return None
*/
void CheckUp::Check_PushButton(MiniRex* Robot)
{
    char PB;
    PB=Robot->Firmware.GetButtonStatus();
    if (PB<0) {
        this->DisplayErrorMessageAndValid(Robot,"Push button, error while requesting data");
        return;
    }
    if (PB==1)
        printf ("Push-button status \t\t\t\t\t ERROR \n");
    else
        this->DisplaySuccessMessage("Push-button status");
}




/*!
\brief Check inclinometer data
\return None
*/
void CheckUp::Check_Inclinometer(MiniRex* Robot)
{
    float Roll,Pitch;
    char Buffer[128];
    char Ret=Robot->Firmware.GetRollPitch(&Roll,&Pitch);
    if (Ret!=1) {
        this->DisplayErrorMessageAndValid(Robot,"Roll and pitch, error while requesting data");
        return;
    }
    if (Roll<-MAX_ANGLE || Roll>MAX_ANGLE) {
        sprintf (Buffer,"Roll, value out of range %5.3f radians",Roll);
        this->DisplayErrorMessageAndValid(Robot,Buffer);
    }
    else {
        sprintf (Buffer,"Roll %5.3f radians",Roll);
        this->DisplaySuccessMessage(Buffer);
    }

    if (Pitch<-MAX_ANGLE || Pitch>MAX_ANGLE) {
        sprintf (Buffer,"Pitch, value out of range %5.3f radians",Pitch);
        this->DisplayErrorMessageAndValid(Robot,Buffer);
    }
    else {
        sprintf (Buffer,"Pitch %5.3f radians",Pitch);
        this->DisplaySuccessMessage(Buffer);
    }
    fflush(stdout);
}




/*!
\brief Check servomotor
\return None
*/
void CheckUp::Check_Servomotor(MiniRex* Robot)
{
    char Ret=Robot->Firmware.SetServoSpeed(256);
    char Buffer[128];
    if (Ret!=1)
        this->DisplayErrorMessageAndValid(Robot,"Servomotor set speed, error while sending command");
    else
        this->DisplaySuccessMessage("Servomotor set speed");


    // Initialize servomotor at central position

    Ret=Robot->Firmware.SetServoPosition(512);
    if (Ret!=1)
        this->DisplayErrorMessageAndValid(Robot,"Servomotor set position, error while sending command");
    else
        this->DisplaySuccessMessage("Servomotor set position");
    fflush(stdout);


    // Check if the requested position is reached

    unsigned short Position;
    int Time_ms=0;

    do
    {
        Ret=Robot->Firmware.GetServoPosition(&Position);
        if (Ret!=1)
            this->DisplayErrorMessageAndValid (Robot,"Servomotor get position, error while requesting data");

        if (Time_ms>1500)
            this->DisplayErrorMessageAndValid (Robot,"Servomotor motion TimeOut");

        usleep (1000);
        Time_ms++;

    }
    while (Position<510 || Position>514);
    sprintf (Buffer,"Servomotor (Pos=%d)",Position);
    this->DisplaySuccessMessage(Buffer);
    fflush(stdout);
}




/*!
\brief Check tracks actuators
\return None
*/
void CheckUp::Check_Tracks(MiniRex* Robot)
{
    char Buffer[128];

    // Enable the motors

    char Ret=Robot->Firmware.Tracks_Enable();
    if (Ret!=1)
        this->DisplayErrorMessageAndValid(Robot,"Enabling track's motors, error while sending command");
    else
        this->DisplaySuccessMessage("Enabling track's motors");
    fflush(stdout);

    // Check the left motor

    Ret=Robot->Firmware.Tracks_SetVelocity(150,0);
    if (Ret!=1)
        this->DisplayErrorMessageAndValid(Robot,"Set tracks velocity, error while sending command");
    else
        this->DisplaySuccessMessage("Set tracks velocity");
    fflush(stdout);

    // Check encoders for left motor

    int NbIt=0, Left,Right;
    do
    {
        Ret=Robot->Firmware.Tracks_GetPosition(&Left,&Right);
        if (NbIt++>200)
            this->DisplayErrorMessageAndValid(Robot,"Left track motion, TimeOut reached");

        if (Ret!=1)
            this->DisplayErrorMessageAndValid(Robot,"Left track motion, error while requesting data");

        usleep(10000);
    }
    while (Left<4000);
    sprintf (Buffer,"Left track motion (Pos=%d)",Left);
    this->DisplaySuccessMessage(Buffer);

    // Check right motor

    Ret=Robot->Firmware.Tracks_SetVelocity(0,150);
    if (Ret!=1)
        this->DisplayErrorMessageAndValid(Robot,"Set tracks velocity");
    else
        this->DisplaySuccessMessage("Set tracks velocity");
    fflush(stdout);

    // Check encoder for right motor

    do
    {
        Ret=Robot->Firmware.Tracks_GetPosition(&Left,&Right);
        if (NbIt++>200)
            this->DisplayErrorMessageAndValid(Robot,"Right track motion, TimeOut reached");


        if (Ret!=1)
            this->DisplayErrorMessageAndValid(Robot,"Right track motion, error while requesting data");

        usleep(10000);
    }
    while (Right<4000);
    sprintf (Buffer,"Right track motion (Pos=%d)",Left);
    this->DisplaySuccessMessage(Buffer);


    Ret=Robot->Firmware.Tracks_SetVelocity(0,0);
    if (Ret!=1)
        this->DisplayErrorMessageAndValid(Robot,"Stop motors, error while sending command");
    else
        this->DisplaySuccessMessage("Stop motors");
    fflush(stdout);
}




/*!
\brief Check ultrasonic sensors
\return None
*/
void CheckUp::Check_UltraSonic(MiniRex* Robot)
{
    short Range;
    char Buffer[128];
    char Ret=Robot->Firmware.US_FullMeasurement(LEFT_US_SENSOR,&Range);
    if (Ret<0)
        this->DisplayErrorMessageAndValid(Robot,"Getting range for US sensor 1, error while requesting data");
    else {
        sprintf (Buffer,"Getting range for US sensor 1 : %d cm",Range);
        this->DisplaySuccessMessage(Buffer);
    }

    Ret=Robot->Firmware.US_FullMeasurement(CENTER_US_SENSOR,&Range);
    if (Ret<0)
        this->DisplayErrorMessageAndValid(Robot,"Getting range for US sensor 2, error while requesting data");
    else{
        sprintf (Buffer,"Getting range for US sensor 2 : %d cm",Range);
        this->DisplaySuccessMessage(Buffer);
    }

    Ret=Robot->Firmware.US_FullMeasurement(RIGHT_US_SENSOR,&Range);
    if (Ret<0)
        this->DisplayErrorMessageAndValid(Robot,"Getting range for US sensor 3, error while requesting data");
    else{
        sprintf (Buffer,"Getting range for US sensor 3 : %d cm",Range);
        this->DisplaySuccessMessage(Buffer);
    }
    fflush(stdout);
}




/*!
\brief Check available space on disk
\return None
*/
void CheckUp::Check_DiskFreeSpace(MiniRex *Robot)
{
    struct statvfs Stat;
    char Buffer[128];
    if (statvfs("/",&Stat) != 0) {
        this->DisplayErrorMessageAndValid(Robot,"Available space on disk, error while checking disk space");
        return;
    }
    double FreeSpace= ((double)Stat.f_bsize * (double)Stat.f_bavail) / (1024.0*1024.0*1024.0);
    if (FreeSpace<32.0)
    {
        sprintf (Buffer,"Not enough space on disk (%2.1fGo)",FreeSpace);
        this->DisplayErrorMessageAndValid(Robot,Buffer);
    }
    else {
        sprintf (Buffer,"Available space on disk: %2.1fGo",FreeSpace);
        this->DisplaySuccessMessage(Buffer);
    }
}



/*!
\brief Create data directory
\return None
*/
void CheckUp::Check_DataDirectory(MiniRex *Robot)
{
    struct tm * ti;
    time_t rawtime;
    time (&rawtime);
    ti=localtime (&rawtime);
    sprintf (Robot->Data_directory,"%s/MiniRex%d_",Robot->Data_directory,Robot->Robot_ID);
    sprintf (Robot->Data_directory,"%s%d-%d-%d_",Robot->Data_directory,ti->tm_mday,ti->tm_mon,ti->tm_year+1900);
    sprintf (Robot->Data_directory,"%s%d-%d-%d",Robot->Data_directory,ti->tm_hour,ti->tm_min,ti->tm_sec);    
    char Ret=mkdir (Robot->Data_directory,S_IRWXU);
    if (Ret != 0)
        this->DisplayErrorMessageAndValid(Robot,"Data directory creation");
    else
        this->DisplaySuccessMessage("Data directory creation");
}


/*!
\brief Display a success message
\brief [OK] is printed, right aligned, and green
\return None
*/

void CheckUp::DisplaySuccessMessage(const char* Message)
{
    printf ("%s",Message);
    for (int i=strlen(Message)/8;i<9;i++) printf ("\t");
    printf("\x1B[1;32m[OK]\x1B[0m\n" );
    fflush(stdout);
}



void CheckUp::DisplayErrorMessage(const char* Message)
{    
    printf ("%s",Message);
    for (int i=strlen(Message)/8;i<9;i++) printf ("\t");
    printf("\x1B[1;31m[ERROR]\x1B[0m\n" );
    fflush(stdout);
}



/*!
\brief Display the message [ERROR] is printed, right aligned, and red
\brief Wait for error validation using the push button
\return None
*/
void CheckUp::DisplayErrorMessageAndValid(MiniRex* Robot,const char *Message)
{
    char Ret;
    Robot->Kinect.SetLedOption(LED_BLINK_RED_ORANGE);
    this->DisplayErrorMessage(Message);
    printf ("\x1B[s\x1B[0;34m \t\tPress the robot button to ignore the error\x1B[0m \x1B[u");
    fflush(stdout);
    // Wait for button pushed
    do
    {
        Ret=Robot->Firmware.GetButtonStatus();
        if (Ret<0)
        {
            printf ("\nError %d while getting button status\n",Ret);
            fflush(stdout);
        }
    }
    while (Ret!=1);

    // Wait for button released
    do
    {
        Ret=Robot->Firmware.GetButtonStatus();
        if (Ret<0)
        {
            printf ("\nError %d while getting button status\n",Ret);
            fflush(stdout);
        }
    }
    while (Ret!=0);
    printf ("\x1B[K");
    fflush(stdout);
    usleep (200000);
    Robot->Kinect.SetLedOption(LED_OFF);
}



/*!
\brief Check servomotor
\return None
*/
void CheckUp::Check_Kinect(MiniRex* Robot)
{
    char Ret;
    double Angle;
    char Buffer [128];

    // Set tilt angle
    Ret=Robot->Kinect.SetTiltAngle(0);
    if (Ret != 1)
        this->DisplayErrorMessageAndValid(Robot,"Set kinect tilt inclination");
    else
        this->DisplaySuccessMessage("Set kinect tilt inclination");

    // Read tilt angle
    Ret=Robot->Kinect.GetTiltAngle(&Angle);
    if (Ret != 1)
        this->DisplayErrorMessageAndValid(Robot,"Get kinect tilt inclination");
    else
        this->DisplaySuccessMessage("Get kinect tilt inclination");

    // Check tilt angle
    if (Angle<-KINECT_INCLINATION || Angle>KINECT_INCLINATION) {
        sprintf (Buffer,"Kinect inclination value out of range %5.3f",Angle);
        this->DisplayErrorMessageAndValid(Robot,Buffer);
    }
    else {
        sprintf (Buffer,"Kinect inclination %5.3f",Angle);
        this->DisplaySuccessMessage(Buffer);
    }

}
