#include "KinectClass.h"

// VID and PID for Kinect and motor/acc/leds
#define MS_MAGIC_VENDOR 0x45e
#define MS_MAGIC_MOTOR_PRODUCT 0x02b0



KinectClass::KinectClass()
{
    KinectConnected=false;
    ImageWidth=0;
    ImageHeight=0;
    DepthWidth=0;
    DepthHeight=0;
}

KinectClass::~KinectClass()
{
    if (KinectConnected) {        
        this->Close();
        printf ("Close Kinect connection\n");
    }
}


/*!
     \brief Open and connect to the Kinect device
     \return 1 success
     \return -1 error while allocating memory for the error messages
     \return -2 error while loading configuration file
     \return -3 error while finding depth sensor node
     \return -4 error while finding image sensor node
     \return -5 error while changing point of view of the depth map
     \return -6 error while connecting to the auxilery devices (Motor, ...)
     \return -7 error while turing off the led
  */
char KinectClass::Init(const char *XML_Config_File_Path)
{
    XnStatus nRetVal = XN_STATUS_OK;
    XnEnumerationErrors* pErrors;

    nRetVal = xnEnumerationErrorsAllocate(&pErrors);
    if (nRetVal != XN_STATUS_OK) return -1;

    nRetVal = xnInitFromXmlFileEx(XML_Config_File_Path, &pContext, pErrors, &hScriptNode);
    if (nRetVal != XN_STATUS_OK) return -2;

    nRetVal = xnFindExistingRefNodeByType(pContext, XN_NODE_TYPE_DEPTH, &hDepth);
    if (nRetVal != XN_STATUS_OK) return -3;

    nRetVal = xnFindExistingRefNodeByType(pContext, XN_NODE_TYPE_IMAGE, &hImage);
    if (nRetVal != XN_STATUS_OK) return -4;

    pDepthMD = xnAllocateDepthMetaData();
    pImageMD = xnAllocateImageMetaData();

    nRetVal=xnSetViewPoint(hDepth,hImage);
    if (nRetVal != XN_STATUS_OK) return -5;

    KinectConnected=true;

    if (this->OpenAuxDevice()!=1) return -6;

    if (this->SetLedOption(LED_OFF)!=1) return -7;
    return 1;
}


/*!
     \brief Free memory and release kinect nodes
  */
void KinectClass::Close()
{
    xnFreeDepthMetaData(pDepthMD);
    xnFreeImageMetaData(pImageMD);
    xnProductionNodeRelease(hDepth);
    xnProductionNodeRelease(hImage);
    xnProductionNodeRelease(hScriptNode);
    xnContextRelease(pContext);
    KinectConnected=false;
    this->CloseAuxDevice();
}



/*!
     \brief Acquire depth and image map from the Kinect
     \brief Each time the function is called, the number of acquisition
     \brief is incremented
     \return 1 success
     \return -1 error while requesting data
*/
char KinectClass::Acquire()
{
    XnStatus nRetVal = XN_STATUS_OK;

    // For for synchro between sensors
    nRetVal = xnWaitAndUpdateAll(pContext);
    if (nRetVal != XN_STATUS_OK) return -1;

    // Get meta data for the depth map
    xnGetDepthMetaData(hDepth, pDepthMD);
    pDepth = pDepthMD->pData;
    DepthWidth = pDepthMD->pMap->Res.X;
    DepthHeight = pDepthMD->pMap->Res.Y;


    // Get meta data for the image
    xnGetImageMetaData(hImage, pImageMD);
    pImage = pImageMD->pData;
    ImageWidth = pImageMD->pMap->Res.X;
    ImageHeight = pImageMD->pMap->Res.Y;

    return 1;
}


/*!
     \brief Convert data from kinect in 3D colored cloud points
     \return number of points in the cloud
  */
void KinectClass::ConvertToCloud()
{
    NbPointCloud=0;
    for (int y=0;y<480;y++)
    {
        for (int x=0;x<640;x++)
        {
            if (pDepth[DepthWidth*y + x]!=0)
            {
                PointsCloud[NbPointCloud].Z = pDepth[DepthWidth*y + x];
                PointsCloud[NbPointCloud].X = x;
                PointsCloud[NbPointCloud].Y = y;
                ColorsCloud[NbPointCloud].nRed   = pImage[(ImageWidth*y + x)*3];
                ColorsCloud[NbPointCloud].nGreen = pImage[(ImageWidth*y + x)*3 +1];
                ColorsCloud[NbPointCloud].nBlue  = pImage[(ImageWidth*y + x)*3 +2];
                NbPointCloud++;
            }
        }
    }
    xnConvertProjectiveToRealWorld(hDepth,640*480,PointsCloud,PointsCloud);    
}


/*!
     \brief Save the image from the last acquisition in a BMP file
     \return 1 success
     \return -1 error while saving BMP file
  */
char KinectClass::SaveImage(const char *FileName)
{

    // Convert in QImage and save the raw map
    QImage *Img = new QImage( pImage,ImageWidth,ImageHeight,QImage::Format_RGB888);

    if (!Img->save(FileName,"BMP",100))
        return -1;
    return 1;
}


/*!
     \brief Save the depth map from the last acquisition in file
     \return 1 success
     \return -1 error while opening file
     \return -2 error while saving width
     \return -3 error while saving height
     \return -4 error while saving data
  */
char KinectClass::SaveDepth(const char* FileName)
{
    FILE *fd;
    fd = fopen (FileName,"w");
    if (fd==NULL) return -3;
    char Buffer[10]="DEPTHRAW_";
    if (!fwrite (Buffer,8,1,fd)) return -4;
    if (!fwrite (&DepthWidth,sizeof(XnUInt32),1,fd)) return -4;
    if (!fwrite (&DepthHeight,sizeof(XnUInt32),1,fd)) return -4;
    if (!fwrite (pDepth,sizeof(XnDepthPixel)*DepthWidth*DepthHeight,1,fd)) return -4;
    fclose (fd);
    return 1;
}


/*!
     \brief Save the 3D point cloud from the last acquisition in file
            ConvertToCloud must be called previously
     \return 1 success
     \return -1 error while opening file
     \return -2 error while saving width
     \return -3 error while saving height
     \return -4 error while saving data
  */
char KinectClass::SaveCloud(const char* FileName)
{
    FILE *fd;
    fd = fopen (FileName,"w");
    if (fd==NULL) return -3;
    char Buffer[10]="CLOUD3D_";
    if (!fwrite (Buffer,8,1,fd)) return -4;
    if (!fwrite (&NbPointCloud,sizeof(int),1,fd)) return -4;
    if (!fwrite (&PointsCloud,NbPointCloud * sizeof(XnPoint3D),1,fd)) return -4;
    if (!fwrite (&ColorsCloud,NbPointCloud * sizeof(XnRGB24Pixel),1,fd)) return -4;
    fclose (fd);
    return 1;
}



/*!
     \brief Open auxilary device (Motor, Tilt and Led)
     \return 1 success
     \return -1 failed to initialize libusb
     \return -2 error while getting device list
     \return -3 device was found, but an error occured while trying to open the device
     \return -4 device was not found
  */
int KinectClass::OpenAuxDevice()
{    
    dev = NULL;
    devs = NULL;
    if (libusb_init(0)) return -1;
    ssize_t cnt = libusb_get_device_list (0, &devs); //get the list of devices
    if (cnt < 0) return -2;

    for (int i = 0; i < cnt; i++)
    {
        struct libusb_device_descriptor desc;
        const int r = libusb_get_device_descriptor (devs[i], &desc);
        if (r < 0) continue;
        // Search for the aux
        if (desc.idVendor == MS_MAGIC_VENDOR && desc.idProduct == MS_MAGIC_MOTOR_PRODUCT)
        {
            if ((libusb_open (devs[i], &dev) != 0) || (dev == 0)) return -3;            
            libusb_claim_interface (dev, 0);                    // Device is found, we claim the device
            return 1;
        }
    }
    return -4;
}


/*!
     \brief Close auxilary device (Motor, Tilt and Led)
*/
void KinectClass::CloseAuxDevice()
{
    if (dev) libusb_release_interface(dev, 0);                      // Release the device
    if (devs) libusb_free_device_list (devs, 1);                    // Free the list, unref the devices in it
}

/*!
     \brief Get tilt angle
     \return 1 success
     \return -1 no device initialized
     \return -2 data has not been read successfully from the USB device
  */
char KinectClass::GetTiltAngle(double *TiltAngle)
{
    if (!dev) return -1;
    uint8_t buf[10];
    const int ret = libusb_control_transfer(dev, 0xC0, 0x32, 0x0, 0x0, buf, 10, 0);
    if (ret != 10) return -2;
    const int8_t tilt_angle = (int8_t)buf[8];
    const uint8_t tilt_status = buf[9];    
    *TiltAngle=double(tilt_angle) / 2.;
    return 1;
}

/*!
     \brief Get tilt status
     \return >0 success (current status)
     \return -1 no device initialized
     \return -2 data has not been read successfully from the USB device
  */
char KinectClass::GetTiltStatus()
{
    if (!dev) return -1;
    uint8_t buf[10];
    const int ret = libusb_control_transfer(dev, 0xC0, 0x32, 0x0, 0x0, buf, 10, 0);
    if (ret != 10) return -2;
    return buf[9];
}


/*!
     \brief Set tilt angle of the Kinect
     \return 1 success
     \return -1 no device initialized
     \return -2 data has not been write successfully to the USB device
     \return -3 TimeOut while waiting the end of motion
     \return -4 Motor limit reached
  */
char KinectClass::SetTiltAngle(const double value)
{
    if (!dev) return -1;
    uint8_t empty[0x1];
    double angle = value;
    angle = (angle<MIN_TILT_ANGLE) ? MIN_TILT_ANGLE : ((angle>MAX_TILT_ANGLE) ? MAX_TILT_ANGLE : angle);
    angle = angle * 2;
    if (libusb_control_transfer(dev, 0x40, 0x31, (uint16_t)angle, 0x0, empty, 0x0, 0)) return -2;
    int TimeOut=0;
    char Status;
    do {
        usleep(1000);
        TimeOut++;
        if (TimeOut==10000) return -3;
        Status=this->GetTiltStatus();
        if (Status==TILT_STATUS_LIMIT) return -4;
    } while (Status!=TILT_STATUS_STOPPED);
    return 1;
}



/*!
     \brief Set the Kinect led status
     \return 1 success
     \return -1 no device initialized
     \return -2 data has not been write successfully to the USB device
  */
int KinectClass::SetLedOption(const uint16_t value)
{
    if (!dev) return -1;
    uint8_t empty[1];
    const uint16_t option = value;
    if (libusb_control_transfer(dev, 0x40, 0x06, (uint16_t)option, 0x0, empty, 0x0, 0)) return -2;
    return 1;
}
