#include <Lidar_Hokuyo.h>
#include <termios.h>
#include <fcntl.h>   											// File control definitions
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>



Lidar_Hokuyo::Lidar_Hokuyo()
{
    LidarConnected=false;
}

Lidar_Hokuyo::~Lidar_Hokuyo()
{
    if (LidarConnected) {
        this->Close();
        printf ("Close Lidar connection\n");
    }
}


// Function open
// Open the URG-04LX device 
// Connect in mode SPIC2.0 at 115200 bauds
// Output 	: 1 Sucess
//		: -1 unable to open device
//              : -2 failed to swithc in SCIP2.0
//		: -3 Failed to turn on the laser
char Lidar_Hokuyo::Open(const char DevicePort[])
{

        struct termios options;                                         // Structure with the device's options
        int Ret;							// Used for returned functions values


        fd = open(DevicePort, O_RDWR | O_NOCTTY | O_NDELAY);            // Open port
        if (fd == -1) return fd;                                        // If the device is not open, return -1
        fcntl(fd, F_SETFL, FNDELAY);                                    // Open the device in nonblocking mode

        tcgetattr(fd, &options);											// Get the current options of the port
	bzero(&options, sizeof(options));									// Clear all the options
    	cfsetispeed(&options, B115200);										// Set the baud rate at 115200 bauds
    	cfsetospeed(&options, B115200);
	options.c_cflag |= ( CLOCAL | CREAD |  CS8);						// Configure the device : 8 bits, no parity, no control
	options.c_iflag |= ( IGNPAR | IGNBRK );
	options.c_cc[VTIME]=0;												// Timer unused
	options.c_cc[VMIN]=0;												// At least on character before satisfy reading
    	tcsetattr(fd, TCSANOW, &options);									// Activate the settings
	
        Ret=Switch_SCIP2();
        if (Ret!=0)	return -2;												// Check if the laser is turned on
        Ret=Laser_On();												// Turn on the laser
        if (Ret!=0)	return -3;												// Check if the laser is turned on
        LidarConnected=true;
        return (1);
}



// Function GetDistance
// Input :	fd : file descriptor
// Output :	>0 : TimeStamp (Time in milliseconds since the sensor has been turned on)
//		-1 : failed to write the command on the device
//		-2 : Wrong data received from the sensor
//		-3 : Status is not 00
//		-4 : Wrong checksum
//		-5 : Distance higher than 50000, bug ???
int Lidar_Hokuyo::GetDistance (int Start,int End,int Cluster, int *Distance)
{
	char 	Cmd[14];													// Command to send
	char 	Buffer[67];													// Buffer for reading data from the sensor
	char 	DataDec[3];													// Temporary array for decoding the data
	char 	Status=0;													// Status = the current character in the three decoding process
	int	NbBytes,i;													// Number of bytes read from the sensor
	int	TimeStamp;													// Time of the acquisition
	int 	Ret;
	
	
	sprintf (Cmd,"GD%4d%4d%2d\n",Start,End,Cluster);
        for (i=0;i<(int)strlen(Cmd);i++)
		if (Cmd[i]==' ') Cmd[i]='0';
	tcflush(fd, TCIFLUSH);												// Empty the file descriptor
	NbBytes = write(fd,Cmd ,strlen(Cmd));								// Send the command on the sensor
	if (NbBytes<0)	return -1;											// Failed to send command ?

        Read (Buffer, sizeof(Buffer),200);							// The sensor must return the same command
	if (strcmp(Buffer,Cmd))	return -2;									// Error while receiving the command
        Read (Buffer, sizeof(Buffer),200);							// The sensor must return Status = 00
	if (strcmp(Buffer,"00P\n"))	return -3;								// Error : Status is not 00
        Ret=Read (Buffer, sizeof(Buffer),200);						// Time Stamp
        if (CheckSum(Buffer)) return -4;								// CheckSum
        TimeStamp=Decode4c(Buffer);										// Store the timestamp
	

        while ((Ret=Read (Buffer, sizeof(Buffer),2000))==0)			// Read a string from fd with a TimeOut of 200ms
	{
		if (Buffer[0]=='\n') {										// End of data acquisition
			*Distance=-1;
			return TimeStamp;
		}

                if (CheckSum(Buffer)) return -4;						// CheckSum

		NbBytes=strlen(Buffer)-2;									// Calculate the number of usefull bytes
		for (i=0;i<NbBytes;i++) {									// For each byte
                        DataDec[(int)Status++]=Buffer[i];   					// Store the curent byte
			if (Status==3) {										// Three bytes are available
                                *Distance=Decode3c (DataDec);					// Decode the bytes
				Distance++;											// Next Data
				Status=0;											// Reset Status
			}
		}
	}
	return -2;
}


// Function GetDistanceIntensity
// Input :	fd : file descriptor
// Output :	>0 : TimeStamp (Time in milliseconds since the sensor has been turned on)
//		-1 : failed to write the command on the device
//		-2 : Wrong data received from the sensor
//		-3 : Status is not 00
//		-4 : Wrong checksum
//		-5 : Distance higher than 8000, bug ???
int Lidar_Hokuyo::GetDistanceIntensity(int Start,int End, int *Distance, int *Intensity)
 {
	char 	Cmd[14];													// Command to send
	char 	Buffer[67];													// Buffer for reading data from the sensor
	char 	DataDec[3];													// Temporary array for decoding the data
	char 	Status=0;													// Status = the current byte in the three decoding process
	char	DtaInt=0;													// Data or Intensity Data
	int	NbBytes,i;													// Number of bytes read from the sensor
	int	TimeStamp;													// Time of the acquisition
	int 	Ret;
	
	
	sprintf (Cmd,"GD%4d%4dFF\n",Start,End);
        for (i=0;i<(int)strlen(Cmd);i++)
		if (Cmd[i]==' ') Cmd[i]='0';
        tcflush(fd, TCIFLUSH);							// Empty the file descriptor
        NbBytes = write(fd,Cmd ,strlen(Cmd));					// Send the command on the sensor
        if (NbBytes<0)	return -1;						// Failed to send command ?
#ifdef ___DEBUG__
        printf ("    GetDataIntensity : Command send : %s\n",Cmd);
#endif
        Read (Buffer, sizeof(Buffer),200);				// The sensor must return the same command
        if (strcmp(Buffer,Cmd))	return -2;					// Error while receiving the command
        Read (Buffer, sizeof(Buffer),200);				// The sensor must return Status = 00
        if (strcmp(Buffer,"00P\n"))	return -3;				// Error : Status is not 00
        Ret=Read (Buffer, sizeof(Buffer),200);				// Time Stamp
#ifdef ___DEBUG__
        if (CheckSum(Buffer)) {
                printf ("    GetDataIntensity : Checksum ERROR lenght = %d Ret=%d\n",strlen(Buffer),Ret);
                printf ("    GetDataIntensity : Buffer = %s",Buffer);
	}
#endif
        if (CheckSum(Buffer)) return -4;					// CheckSum
        TimeStamp=Decode4c(Buffer);						// Store the timestamp
	

        while ((Ret=Read (Buffer, sizeof(Buffer),2000))==0)		// Read a string from fd with a TimeOut of 200ms
	{
                if (Buffer[0]=='\n') {						// End of data acquisition
                        *Distance=-1;						// Mark the end of the array
                        *Intensity=-1;						// Mark the end of the array
			return TimeStamp;
		}
#ifdef ___DEBUG__
        if (CheckSum(Buffer)) {
                printf ("    GetDataIntensity : Checksum ERROR lenght = %d Ret=%d\n",strlen(Buffer),Ret);
                printf ("    GetDataIntensity : Buffer = %s",Buffer);
	}
#endif

                if (CheckSum(Buffer)) return -4;				// CheckSum

                NbBytes=strlen(Buffer)-2;					// Calculate the number of usefull bytes

                for (i=0;i<NbBytes;i++) {					// For each byte
                        DataDec[(int)Status++]=Buffer[i];			// Store the curent byte
                        if (Status==3 && DtaInt==0) {				// Three bytes are available, this is the distance
                                *Distance=Decode3c (DataDec);		// Decode the bytes
				Distance++;											// Next Data
				Status=0;											// Reset Status
				DtaInt=1;											// Next data is intensity
			}
                        if (Status==3 && DtaInt==1) {				// Three bytes are available, this is the intensity
                                *Intensity=Decode3c (DataDec);		// Decode the bytes
				Intensity++;											// Next Data
				Status=0;											// Reset Status
				DtaInt=0;											// Next data is a distance
			}

		}
	}
	return -2;
}


// Function GetDistanceIntensityAGC
// Input :	fd : file descriptor
// Output :	>0 : TimeStamp (Time in milliseconds since the sensor has been turned on)
//		-1 : failed to write the command on the device
//		-2 : Wrong data received from the sensor
//		-3 : Status is not 00
//		-4 : Wrong checksum
//		-5 : Distance higher than 8000, bug ???
int Lidar_Hokuyo::GetDistanceIntensityAGC(int Start,int End, int *Distance, int *Intensity,int *AGC)
 {
	char 	Cmd[14];													// Command to send
	char 	Buffer[67];													// Buffer for reading data from the sensor
	char 	DataDec[3];													// Temporary array for decoding the data
	char 	Status=0;													// Status = the current byte in the three decoding process
	char	DtaInt=0;													// Data or Intensity Data
	int	NbBytes,i;													// Number of bytes read from the sensor
	int	TimeStamp;													// Time of the acquisition
	int 	Ret;														// Returned value of the function called
	
	
	sprintf (Cmd,"GD%4d%4dDF\n",Start,End);
        for (i=0;i<(int)strlen(Cmd);i++)
		if (Cmd[i]==' ') Cmd[i]='0';
	tcflush(fd, TCIFLUSH);												// Empty the file descriptor
	NbBytes = write(fd,Cmd ,strlen(Cmd));								// Send the command on the sensor
	if (NbBytes<0)	return -1;											// Failed to send command ?
        Read (Buffer, sizeof(Buffer),200);							// The sensor must return the same command
	if (strcmp(Buffer,Cmd))	return -2;									// Error while receiving the command
        Read (Buffer, sizeof(Buffer),200);							// The sensor must return Status = 00
	if (strcmp(Buffer,"00P\n"))	return -3;								// Error : Status is not 00
        Ret=Read (Buffer, sizeof(Buffer),200);						// Time Stamp
        if (CheckSum(Buffer)) return -4;                                                // CheckSum
        TimeStamp=Decode4c(Buffer);                                                     // Store the timestamp
	

        while ((Ret=Read (Buffer, sizeof(Buffer),2000))==0)                         // Read a string from fd with a TimeOut of 200ms
	{
                if (Buffer[0]=='\n') {                                                      // End of data acquisition
                        *Distance=-1;                                                       // Mark the end of the array
                        *Intensity=-1;                                                      // Mark the end of the array
			*AGC=-1;												// Mark the end of the array
			return TimeStamp;
		}

                if (CheckSum(Buffer)) return -4;						// CheckSum

                NbBytes=strlen(Buffer)-2;                                       		// Calculate the number of usefull bytes

                for (i=0;i<NbBytes;i++) {							// For each byte
                        DataDec[(int)Status++]=Buffer[i];						// Store the curent byte
                        if (Status==3 && DtaInt==0) {						// Three bytes are available, this is the Distance
                                *Distance=Decode3c (DataDec);				// Decode the bytes
				Distance++;											// Next Data
				Status=0;											// Reset Status
				DtaInt=1;											// Next data is intensity
			}
                        if (Status==3 && DtaInt==1) {						// Three bytes are available, this is the Intensity
                                *Intensity=Decode3c (DataDec);				// Decode the bytes
				Intensity++;										// Next Data
				Status=0;											// Reset Status
				DtaInt=2;											// Next data is a distance
			}
                        if (Status==3 && DtaInt==2) {						// Three bytes are available, this is the AGC
                                *AGC=Decode3c (DataDec);					// Decode the bytes
				AGC++;												// Next Data
				Status=0;											// Reset Status
				DtaInt=0;											// Next data is a distance
			}
		}
	}
	return -2;
}



// Function Laser_On
// Turn on the laser
// Input 	: file descriptor
// Output	: 0 success
//		: -1 failed to write the command on the device
//		: -2 Wrong data received from the sensor
//		: -3 Unable to control due to laser malfunction
int Lidar_Hokuyo::Laser_On ()
{
	char 	Buffer[5];													// Buffer for reading data from the sensor
	int	NbBytes;													// Number of bytes read from the sensor

	tcflush(fd, TCIFLUSH);												// Empty the file descriptor
	NbBytes = write(fd, "BM\n", 3);										// Send the command on the sensor
	if (NbBytes<0)	return -1;											// Failed to send command ?
#ifdef ___DEBUG__
        printf ("    Laser_On : Command send                                            [OK]\n");
#endif
        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
	if (strcmp(Buffer,"BM\n")) return -2;								// The return string is not in accordance with BM[LF]
#ifdef ___DEBUG__
        printf ("    Laser_On : BM[LF]                                                  [OK]\n");
#endif

        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
	if (!strcmp(Buffer,"01Q\n")) return -3;								// Laser malfunction ?
	if (strcmp(Buffer,"00P\n") && strcmp(Buffer,"02R\n")) return -2;	// The return string is not in accordance with OxS[LF]
        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
        if (strcmp(Buffer,"\n")) return -2;									// The return string is not in accordance with [LF]
	return 0;
}

// Function Laser_Off
// Turn off the laser
// Input 	: file descriptor
// Output	: 0 success
//		: -1 failed to write the command on the device
//		: -2 Wrong data received from the sensor
int Lidar_Hokuyo::Laser_Off ()
{
        char 	Buffer[5];													// Buffer for reading data from the sensor
        int	NbBytes;													// Number of bytes read from the sensor

        tcflush(fd, TCIFLUSH);												// Empty the file descriptor
        NbBytes = write(fd, "QT\n", 3);										// Send the command on the sensor
        if (NbBytes<0)	return -1;											// Failed to send command ?
        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
        if (strcmp(Buffer,"QT\n")) return -2;								// The return string is not in accordance with BM[LF]
        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
        if (strcmp(Buffer,"\n")) return -2;									// The return string is not in accordance with [LF]
        return 0;
}



// Function Switch_SCIP2
// Switch the sensor to SCIP 2.0
// Input 	: file descriptor
// Output	: 0 success or already in SCIP2.0
//		: -1 failed to write the command on the device
//		: -2 Echo not received from the sensor
//              : -3 Unknow error received form the sensor
//              : -4 Line feed is missing
int Lidar_Hokuyo::Switch_SCIP2 ()
{
        char 	Buffer[9];													// Buffer for reading data from the sensor
        int	NbBytes;													// Number of bytes read from the sensor

        tcflush(fd, TCIFLUSH);												// Empty the file descriptor
        NbBytes = write(fd, "SCIP2.0\n", 8);										// Send the command on the sensor
        if (NbBytes<0)	return -1;											// Failed to send command ?
        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
        if (strcmp(Buffer,"SCIP2.0\n")) return -2;								// The return string is not in accordance with BM[LF]
        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
        if (strcmp(Buffer,"0Ee\n") && (strcmp(Buffer,"0\n"))) return -3;
        Read (Buffer, sizeof(Buffer),200);							// Read a string from fd with a TimeOut of 200ms
        if (strcmp(Buffer,"\n")) return -4;									// The return string is not in accordance with [LF]

        return 0;
}


// Function Close
// Turn off the laser
// Close the file descriptor on the device
// input : 	fd :file descriptor
// output : 	0 : Sucess
//		<0 : error while turning off the laser
int Lidar_Hokuyo::Close()
{
        int Ret=Laser_Off();
	close (fd);
        LidarConnected=false;
	return Ret;
}




// Function Read
// Read a string from the device terminated by line feed (LF (0x0A))
// Input :	fd : File descriptor
//		Buffer : pointer on the buffer (for returning th string)
//		Buf_Size : maximum number of character to read
//		TimeOut : After this delay in millisecond, the reading is aborded
// Output :	0 : read success
//		-1 : maximum number of character reached
//		-2 : TimeOut reached
//		-3 : End of buffer reached before LF
int Lidar_Hokuyo::Read (char *Buffer, int Buf_Size, long int TimeOut)
{
	int	NbBytes;													// Number of bytes read from the sensor

        this->TimeOut(1);														// Initialize the timer for the TimeOut
	do {
		NbBytes=read(fd,Buffer,1);									// Read one character on the buffer
		if (NbBytes>0) {											// A byte has been received
			Buf_Size--;												// Decrement the size of the buffer
			if (*Buffer=='\n')	{									// If it is line feed
				Buffer++;	
				*Buffer='\0';										// Add the end of string char
				return 0;											// Sucess : return 0
			}
			Buffer++;												// Next character
			if (Buf_Size==1) {										// The buffer is full
				*Buffer='\0';										// Add the end of string char
				return -3;											// Sucess : return 0
			}			
		}
	}	
        while (this->TimeOut(0)<TimeOut);										// while the TimeOut is not reached
	*Buffer='\0';
	return -2;
}

// Function CheckSum
// Return 0 if the CheckSum is in accordance with Hokuyo's specifications
// Buffer must be as follow : [DATA nbytes] [CheckSUm 1byte] [LF 1byte]
// Input : 		Buffer : string to anaylse
// Output :		0 : the checksum is OK
// 			-1 : there is a problem !
char Lidar_Hokuyo::CheckSum(char *Buffer)
{
	int Sum=0;															// Used to compute the checksum
	unsigned char CheckSum=0;											// Checksum read from buffer
	while (*Buffer!='\n') {												// While the chararcter in buffer is not LF
		Sum+=*Buffer;												// Add the byte to the sum
		Buffer++;													// Next byte
	}
	Buffer--;															// Go back the the checksum
	CheckSum=*Buffer;													// Memorize the checksum in buffer
	Sum-=CheckSum;														// Substract the checksum from Sum
	Sum=0x30+(Sum&0x3F);												// Compute the checksum
	if (Sum==CheckSum) return 0;										// Compare and return 0 if it is the same
	else return -1;														// Check Sum is wrong
}

// Function Decode4c
// Decode Buffer according to Four Character Encoding
// Input :	Buffer : String to decode
// Output :	Integer decoded
int Lidar_Hokuyo::Decode4c (char *Buffer)
{
	int Value=Buffer[3]-0x30;
	Value+=(int)(Buffer[2]-0x30)<<6;
	Value+=(int)(Buffer[1]-0x30)<<12;
	Value+=(int)(Buffer[0]-0x30)<<18;
	return Value;
}



// Function Decode3c
// Decode Buffer according to Three Character Encoding
// Input :	Buffer : String to decode
// Output :	Integer decoded
int Lidar_Hokuyo::Decode3c (char *Buffer)
{
	int Value=Buffer[2]-0x30;
	Value+=(int)(Buffer[1]-0x30)<<6;
	Value+=(int)(Buffer[0]-0x30)<<12;
	return Value;
}

// Function Time_Out
// Return the elapsed time in millisecond since the last call with Status=1
// Input :	Status = 0 : return ellapsed time
//		Status = 1 : memorize current time
// Output :	Elapsed time since last call with Status=1
long int Lidar_Hokuyo::TimeOut(char Status)
{
	static struct timeval PreviousTime={0,0};							// Memorize time of previous call
	struct timeval CurrentTime;											// Current time
	long int sec,usec;

	if (Status==1) {													// Request a memorization
		gettimeofday(&PreviousTime, NULL);							// Memorize current time
		return 0;													// Return 0
	}

	gettimeofday(&CurrentTime, NULL);									// Get current time

	sec=CurrentTime.tv_sec-PreviousTime.tv_sec;							// Compute the number of second elapsed since last call
	usec=CurrentTime.tv_usec-PreviousTime.tv_usec;						// Compute
	if (usec<0) {														// If the previous usec is higher than the current one
		usec=1000000-PreviousTime.tv_usec+CurrentTime.tv_usec;		// Recompute the microseonds
		sec--;														// Substract one second
	}
	return sec*1000+usec/1000;
}




